#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "engine.h"
extern "C" { void draw_tile(int number,int x,int y); }

struct Tile tile[256];
int tileCount;
int charToTile[256];
int editMode;

#define FACING_RIGHT 0
#define FACING_LEFT 1

//variables
struct SpriteDescription marioWalkDesc={
	4, { {0,0},{1,0},{2,0},{1,0}, }, 4
};
struct SpriteDescription marioSwimDesc={
	4, { {7,0},{8,0},{9,0},{10,0} }, 4
};
//enemy
struct SpriteDescription koopaWalkDesc={
	2, { {22,0},{23,0} }, 15
};

struct Sprite mario;
struct Sprite koopa[32];
struct Map map;

//function declarations
void DrawMapTile(int screenX,int screenY,int tileX,int tileY);
void LoadMap(const char *filename);

void EngineInit()
{
	if(tileCount>0) return;
	
	char buffer[256];
	FILE *file=fopen("Scenery.txt","r");

	printf("Loading in engine settings....\n");
	if(!file) {
		printf("No scenery.txt config file.  That's no fun\n");
	}
	int tagActive=0;
	int frame=0;
	while(fgets(buffer,255,file)) {
		buffer[255]=0;
		if(strchr(buffer,'\r')) strchr(buffer,'\r')[0]=0;
		if(strchr(buffer,'\n')) strchr(buffer,'\n')[0]=0;
		if(buffer[0]=='/') continue;	// skip comments.
		if(buffer[0]=='[') {
			// new tag type
			//printf("Tag count: %d.  Last record: '%c' name: '%s'\n",tileCount, tile[tileCount].shortForm, tile[tileCount].name);
			if(tagActive) tileCount++;
			frame=0;
			tagActive=1;
			if(strchr(buffer,']')) strchr(buffer,']')[0]=0;
			strcpy(tile[tileCount].name,buffer+1);
		} else if(buffer[0]=='{') {
			sscanf(buffer,"{Frame.%d}",&frame);
			if(frame>tile[tileCount].frameCount) {
				tile[tileCount].frameCount=frame;
			}
		} else if(strchr(buffer,'=')) {
			char *name,*value;

			name=buffer;
			value=strchr(buffer,'=');
			value[-1]=0;	// trim a space
			value++;
			value++;
			if(strcmp(name,"ShortForm")==0) {
				tile[tileCount].shortForm=value[0];
				charToTile[(unsigned)value[0]]=tileCount;
			} else if(strcmp(name,"Over")==0) {
				sscanf(value,"%d",&tile[tileCount].x);
				if(frame>0) {
					tile[tileCount].frame[frame-1].x=tile[tileCount].x;
				}
			} else if(strcmp(name,"Down")==0) {
				sscanf(value,"%d",&tile[tileCount].y);
				if(frame>0) {
					tile[tileCount].frame[frame-1].y=tile[tileCount].y;
				}
			} else if(strcmp(name,"Width")==0) {
				sscanf(value,"%d",&tile[tileCount].w);
				if(frame>0) {
					tile[tileCount].frame[frame-1].w=tile[tileCount].w;
				}
			} else if(strcmp(name,"Height")==0) {
				sscanf(value,"%d",&tile[tileCount].h);
				if(frame>0) {
					tile[tileCount].frame[frame-1].h=tile[tileCount].h;
				}
			} else if(strcmp(name,"Type")==0) {
				if(strcmp(value,"SolidGround")==0) {
					tile[tileCount].type=TT_SOLIDGROUND;
				} else if(strcmp(value,"BackGround")==0) {
					tile[tileCount].type=TT_BACKGROUND;
				} else if(strcmp(value,"ItemBlock")==0) {
					tile[tileCount].type=TT_ITEMBLOCK;
				} else if(strcmp(value,"BumpableBlock")==0) {
					tile[tileCount].type=TT_BUMPABLEBLOCK;
				} else if(strcmp(value,"Coin")==0) {
					tile[tileCount].type=TT_COIN;
				} else if(strcmp(value,"DangerBlock")==0) {
					tile[tileCount].type=TT_DANGERBLOCK;
				} else if(strcmp(value,"ExitPipe")==0) {
					tile[tileCount].type=TT_EXITPIPE;
				} else if(strcmp(value,"Climbable")==0) {
					tile[tileCount].type=TT_CLIMBABLE;
				} else {
					printf("Found an unknown tile type in the config file: '%s'.",value);
				}
			} else {
				printf("Found an unknown tile attribute name in the config file: '%s'.",name);
			}
		}
		
	}
	fclose(file);
	
	int i;
	for(i=0;i<tileCount;i++) {
		tile[i].w++;
		tile[i].h++;
	}
	
	// Debug
	for(i=0;i<tileCount;i++) {
		if(strcmp(tile[i].name,"Bones")==0) {
			printf("Tile %s: over=%d, down=%d, width=%d, height=%d\n",
			tile[i].name,tile[i].x,tile[i].y,tile[i].w,tile[i].h);
		}
	}
}

// translate a tile in a level file, into a tile in the sprite sheet
int mapCharToTile( char ch)
{
	int id;

	switch(ch) {
	//case ' ': return TILE(29,9);	// space
	case 'u': return ENEM(25,4);	// ghost
	case 'v': return ENEM(7,2);	// bomb
	case 'x': return ENEM(3,3);	// turtle
	case 'y': return ENEM(1,2);	// koopa
	case 'z': return ENEM(20,0);	// bear
	default:
		// still doesn't quite handle multitile objects.
		id=charToTile[(int)ch];
		printf("matched tile '%c'to id %d (%s)\n",ch,id,tile[id].name);
		return TILE(tile[id].x,tile[id].y);
	}
}

void LoadMap(const char *filename)
{
	char line[256];
	printf("Locating map '%s'\n",filename);
	FILE *file = fopen( filename, "r");
	if(!file) return;
	printf("Loading map...\n");
	if(map.map) free(map.map);
	if(map.tile) free(map.tile);
	map.width=30;	// PSP screen is 480/16
	map.height=17;	// PSP screen is 272/16
	map.map = (int *)calloc( sizeof(int), map.width*map.height);
	map.tile = (int *)calloc( sizeof(int), map.width*map.height);
	int j=0;
	int i;
	while( fgets( line, 255, file) ) {
		line[255] = 0;
		if(strrchr( line, '\n') ) strrchr( line, '\n')[0] = 0;
		for(i = 0; i<strlen(line); i++) {
			if( i < map.width) {
				// Set default tile for the map as an open square.
				map.map[i+map.width*j]=TILE(29,9);
				map.tile[i+map.width*j]=0;	// sky
				// Handle special tiles, for active enemies, and Mario's start point.
				switch(line[i]) {
				case 'M':	// mario start point
					//mario.spriteImage=playerImage;
					mario.x=i*16;
					mario.y=j*16;
					mario.facing=FACING_RIGHT;
					mario.health=1;
					mario.speed=0;
					mario.frame=0;
					mario.frameDelay=3;
					mario.w=1;
					mario.h=1;
					mario.desc=&marioWalkDesc;
					mario.vertSpeed=0;
					mario.vertAccel=0;
					break;
				case 'y':	// koopa
					//NewKoopa(i*16,j*16);
					break;
				default:
					{
						int tileNo=mapCharToTile(line[i]);
						int tileId=charToTile[(unsigned char)line[i]];
						map.tile[i+map.width*j]=tileId;
						if(tile[tileId].w>1 || tile[tileId].h>1) {
							if(i>0) {
								int tileLeft=map.map[(i-1)+map.width*j];
								int xx=tileLeft%32;
								int yy=tileLeft/32;
								if(xx>=tile[tileId].x && xx<tile[tileId].x+tile[tileId].w
								&& yy>=tile[tileId].y && yy<tile[tileId].y+tile[tileId].h) {
									int ox=xx-tile[tileId].x;
									int oy=yy-tile[tileId].y;
									ox+=1;
									if(ox>=tile[tileId].w) ox=0;
									tileNo=mapCharToTile(line[i])+ox+oy*TILESACROSS;
								}
							}
							if(j>0) {
								int tileAbove=map.map[i+map.width*(j-1)];
								int xx=tileAbove%32;
								int yy=tileAbove/32;
								if(xx>=tile[tileId].x && xx<tile[tileId].x+tile[tileId].w
								&& yy>=tile[tileId].y && yy<tile[tileId].y+tile[tileId].h) {
									int ox=xx-tile[tileId].x;
									int oy=yy-tile[tileId].y;
									oy+=1;
									if(oy>=tile[tileId].h) oy=0;
									tileNo=mapCharToTile(line[i])+ox+oy*TILESACROSS;
								}
							}
						}
						map.map[i+map.width*j]=tileNo;
					}
				}
			}
		}
		j++;
		if( j >= map.height) break;

	}
	fclose(file);
	printf("Loaded map\n");
}

void DrawMap()
{
	int i,j;
	if(!map.map) return;
	for(j=0; j<SCREENHEIGHT/TILEHEIGHT; j++) {
		for(i=0; i<SCREENWIDTH/TILEWIDTH; i++) {
			int x,y;
			int tileNo=map.map[(i+map.l)+(j+map.t)*map.width];
			if(tileNo==0) tileNo=TILE(29,9);	// great for incomplete maps.
			x=tileNo&31;
			y=(tileNo>>5)&127;
			draw_tile(tileNo,i*16,j*16);
		}
	}
}


void new_game()
{
	EngineInit();
}

void construct_level(int width,int height)
{
	map.map=(int *)calloc(sizeof(int),width*height);
	map.tile=(int *)calloc(sizeof(int),width*height);
	map.width=width;
	map.height=height;
	map.t=0;
	map.l=0;
	map.exitCount=0;
}

void save_level(const char *filename)
{
	FILE *file=fopen(filename,"w");
	int i,j;
	
	for(j=0;j<map.height;j++) {
		for(i=0;i<map.width;i++) {
			fprintf(file,"%c",tile[map.tile[i+j*map.width]].shortForm);
		}
		fprintf(file,"\n");
	}
	fclose(file);
}

void load_level(const char *filename)
{
	LoadMap(filename);	
}
