#ifndef ENGINE_H
#define ENGINE_H

#define TILEWIDTH 16
#define TILEHEIGHT 16
#define PTILEWIDTH 16
#define PTILEHEIGHT 16
#define TILESACROSS 32
#define SCREENWIDTH 480
#define SCREENHEIGHT 272

//definitions
struct SpriteFrame {
	// the column and row of the sprite in the sprite sheet.
	int x,y;	// top left corner is 0,0
};
struct SpriteDescription {
	int frameCount;	// how many there are for this sprite
	struct SpriteFrame frames[8]; // list of pairs of sprites from the sprite sheet
	int delay;	// how long each frame is displayed, in 1/60ths of a second.
};
enum SpriteType { ST_PLAYER, ST_ENEMY, ST_ITEM };
typedef enum SpriteType SpriteType;
enum ModeType { MT_WALK, MT_JUMP, MT_CLIMB, MT_SWIM };
typedef enum ModeType ModeType;
struct Sprite {
	SpriteType type;
	ModeType mode;
	void *spriteImage;	// which image to show.
	int health;	// when 0, this sprite is unused.
	int x,y;	// where is the sprite on screen
	int w,h;	// the size of the sprite in tiles.
	int facing;	// is it facing left or right?
	int speed;	// how fast is it moving per frame.
	int vertSpeed;	// how fast we are going up
	int vertAccel;	// if we are actively jumping at the moment.
	int frame;	// what frame of animation are we showing
	int frameDelay;	// how long until we switch to the next frame
	struct SpriteDescription *desc;	// the list of animation frames, and where they are on the sprite sheet
	int coins;	// how many coins this sprite has collected.
};

enum TileType {
	TT_BACKGROUND, // can walk in front of this
	TT_SOLIDGROUND, // can walk on this; has collision all around
	TT_ITEMBLOCK, // drops powerups
	TT_BUMPABLEBLOCK, // can destroy the block
	TT_COIN, // mooluah
	TT_DANGERBLOCK, // kills on touch
	TT_EXITPIPE, // will exit from the top, but otherwise has collision all around
	TT_CLIMBABLE,	// can climb up this
};
typedef enum TileType TileType;
struct TilePos {
	short x,y,w,h;
};
struct Tile {
	void *spriteImage;	// which sprite sheet it is on.
	char name[128];
	char shortForm;
	TileType type;
	int x,y,w,h;	// units are in tiles.
	int frameCount;	// frame count for animated tiles
	struct TilePos frame[6];
};
extern struct Tile tile[256];
extern int tileCount;
extern int charToTile[256];

struct Map {
	int width,height;
	int t,l;	// top and left
	int *map;
	int *tile;
	int exitCount;
};
#define TILE(x,y) ((x)+(y)*32)
#define ITEM(x,y) ((x)+(y+32)*32)
#define ENEM(x,y) ((x)+(y+64)*32)
#define EXIT(x,y,map) ((x)+(y)*32+map*65536)

extern struct Map map;

//function declarations
void EngineInit();
void DrawMap();
#define draw_map DrawMap
#define tile_to_prop(tileId) (tileId>=0?(&tile[tileId]):NULL)
#define M_SOLIDGROUND 2
#define TileProp Tile
#define game map
void new_game();
void construct_level(int width,int height);
void save_level(const char *filename);
void load_level(const char *filename);
extern int editMode;

#endif
