//OSlib header file
#include <oslib/oslib.h>

//Necessary to create eboot
PSP_MODULE_INFO("Dark Mario", 0, 1, 1);
PSP_MAIN_THREAD_ATTR(THREAD_ATTR_USER | THREAD_ATTR_VFPU);
PSP_HEAP_SIZE_KB(20480);

//declaration of the pointers of our images
OSL_IMAGE *backgroundImage, *itemsImage, *playerImage, *enemiesImage, *sceneryImage, *menuback, *sel;
const char *soundfilename[]={
	"sfx/breakblock.wav", "sfx/clearlevel.wav", "sfx/coinget.wav", "sfx/hitwall.wav", "sfx/jump.wav",
	"sfx/kickshell.wav", "sfx/mariodie.wav", "sfx/mariostart.wav", "sfx/pause.wav", "sfx/powerup.wav",
	"sfx/spitfire.wav", "sfx/stomp.wav"
};
enum SoundId {
	S_BREAKBLOCK, S_CLEARLEVEL, S_COINGET, S_HITWALL, S_JUMP,
	S_KICKSHELL, S_MARIODIE, S_MARIOSTART, S_PAUSE, S_POWERUP,
	S_SPITFIRE, S_STOMP, SOUNDMAX
};
OSL_SOUND *sfx[SOUNDMAX];
OSL_SOUND *music;

int menu = 1; //Determines when the menu is active//
int menusel = 1; //Determines which selection is selected in the menu//
int first_sel_x = 173; //The x position of the first selection//
int first_sel_y = 59; //The y position of the first selection//
int selimagex; //Determines the starting x postion of the selection//
int selimagey; //Determines the starting y postion of the selection//
int num_of_sel = 3;
int x_dis_bet_sel = 0;//The x distance between two selections//
int y_dis_bet_sel = 10;//The y distance between two selections//

#define JUMPACCEL 256
#define MAXVERTSPEED 1792
#define MINVERTSPEED -1792
#define GRAVITY 128

//definitions
struct SpriteFrame {
	// the column and row of the sprite in the sprite sheet.
	int x,y;	// top left corner is 0,0
};
struct SpriteDescription {
	int frameCount;	// how many there are for this sprite
	struct SpriteFrame frames[8]; // list of pairs of sprites from the sprite sheet
	int delay;	// how long each frame is displayed, in 1/60ths of a second.
};
enum SpriteType { ST_PLAYER, ST_ENEMY, ST_ITEM };
typedef enum SpriteType SpriteType;
enum ModeType { MT_WALK, MT_JUMP, MT_CLIMB, MT_SWIM };
typedef enum ModeType ModeType;
struct Sprite {
	SpriteType type;
	ModeType mode;
	OSL_IMAGE *spriteImage;	// which image to show.
	int health;	// when 0, this sprite is unused.
	int x,y;	// where is the sprite on screen
	int w,h;	// the size of the sprite in tiles.
	int facing;	// is it facing left or right?
	int speed;	// how fast is it moving per frame.
	int vertSpeed;	// how fast we are going up
	int vertAccel;	// if we are actively jumping at the moment.
	int frame;	// what frame of animation are we showing
	int frameDelay;	// how long until we switch to the next frame
	struct SpriteDescription *desc;	// the list of animation frames, and where they are on the sprite sheet
	int coins;	// how many coins this sprite has collected.
};

enum TileType {
	TT_BACKGROUND, // can walk in front of this
	TT_SOLIDGROUND, // can walk on this; has collision all around
	TT_ITEMBLOCK, // drops powerups
	TT_BUMPABLEBLOCK, // can destroy the block
	TT_COIN, // mooluah
	TT_DANGERBLOCK, // kills on touch
	TT_EXITPIPE, // will exit from the top, but otherwise has collision all around
	TT_CLIMBABLE,	// can climb up this
};
typedef enum TileType TileType;
struct TilePos {
	short x,y,w,h;
};
struct Tile {
	OSL_IMAGE *spriteImage;	// which sprite sheet it is on.
	char name[128];
	char shortForm;
	TileType type;
	int x,y,w,h;	// units are in tiles.
	int frameCount;	// frame count for animated tiles
	struct TilePos frame[6];
} tile[256];
int tileCount;
int charToTile[256];

#define FACING_RIGHT 0
#define FACING_LEFT 1
struct AnimTile {
	short x,y;
	struct Tile *tile;
	int frame;
};

struct Map {
	int width,height;
	int *map;
	int *tile;
	int exitCount;
	struct AnimTile *animTile;
	int animTileCount;
};
#define TILE(x,y) ((x)+(y)*32)
#define ITEM(x,y) ((x)+(y+32)*32)
#define ENEM(x,y) ((x)+(y+64)*32)
#define EXIT(x,y,map) ((x)+(y)*32+map*65536)

//variables
struct SpriteDescription marioWalkDesc={
	4, { {0,0},{1,0},{2,0},{1,0}, }, 4
};
struct SpriteDescription marioSwimDesc={
	4, { {7,0},{8,0},{9,0},{10,0} }, 4
};
//enemy
struct SpriteDescription koopaWalkDesc={
	2, { {22,0},{23,0} }, 15
};

struct Sprite mario;
struct Sprite koopa[32];
struct Map map;

//function declarations
void ProcessButtons();
void UpdateSprite(struct Sprite *sprite);
void DrawSprite(struct Sprite *sprite);
void DrawMapTile(int screenX,int screenY,int tileX,int tileY);
void NewKoopa(int x,int y);
void LoadMap(const char *filename);
void UpdateMap();
void DrawMap();
void ShowTiles();
void EngineInit();
void playSound( int sound);

int main()
{
	//Initialization of the Oslib library
	oslInit(0);

	//Initialization of the graphics mode
	oslInitGfx(OSL_PF_8888, 1);

	//Sets the transparency color (green)
	oslSetTransparentColor(RGB(0,255,0));

	//loads our images into memory
	backgroundImage = oslLoadImageFile("Background.png", OSL_IN_RAM, OSL_PF_5551);
	if(!backgroundImage) printf("background missing.\n");
	sceneryImage = oslLoadImageFile("Scenery.png", OSL_IN_RAM, OSL_PF_5551);
	if(!sceneryImage) printf("scenery missing.\n");
	enemiesImage = oslLoadImageFile("Enemies.png", OSL_IN_RAM, OSL_PF_5551);
	if(!enemiesImage) printf("enemies missing.\n");
	playerImage = oslLoadImageFile("Player.png", OSL_IN_RAM, OSL_PF_5551);
	if(!playerImage) printf("player missing.\n");
	itemsImage = oslLoadImageFile("Items.png", OSL_IN_RAM, OSL_PF_5551);
	if(!itemsImage) printf("items missing.\n");
	menuback = oslLoadImageFile("BGMenu.png", OSL_IN_RAM, OSL_PF_5551);
	if (!menuback) printf("items missing.\n");
	sel = oslLoadImageFile("sel.png", OSL_IN_RAM, OSL_PF_5551);
	if (!sel) printf("Item Missing.\n");
	


	//Disables the transpaent color (pink)
	oslDisableTransparentColor();

	//verification that all files are present
	if (!backgroundImage || !sceneryImage || !enemiesImage || !playerImage || !itemsImage)
		oslDebug("It is impossible to load one or more of the images that are required in this program. Please make sure you have all the files required in the eboot folder");

	mario.type=ST_PLAYER;
	mario.mode=MT_WALK;
	mario.w=1;
	mario.h=1;
	mario.desc=&marioWalkDesc;

	EngineInit();

	// Debug the Scenery.txt
	oslReadKeys();
	if(osl_keys->held.L) {
		ShowTiles();
	}

    //Sets the sprite's original position on the screen
	LoadMap("map/001.txt");

	//main while loop
	while (!osl_quit) {
		//To be able to draw on the screen
		oslStartDrawing();		
		
		if (menu == 1) {
			//initiate the PSP's buttons
			oslReadKeys();

			oslDrawImage(menuback);
			oslDrawImage(sel);
        
			sel->x = selimagex; //Sets the selection coordinates//
			sel->y = selimagey; //Sets the selection coordinates//
        
			selimagex = first_sel_x+(x_dis_bet_sel*menusel); //Detarmines where the selection image is drawn for each selection//
			selimagey = first_sel_y+(y_dis_bet_sel*menusel); //Detarmines where the selection image is drawn for each selection//
        
			if (osl_keys->pressed.down) menusel++; //Moves the selection down//
			if (osl_keys->pressed.up) menusel--; //Moves the selection up//
        
			if (menusel > num_of_sel) menusel = 1; //Sets the selection back to the top//
			if (menusel < 1) menusel = num_of_sel; //Sets the selection back to the bottom//
			
			if (osl_keys->pressed.cross && menusel == 1) menu=0;
			if (osl_keys->pressed.cross && menusel == 3) osl_quit=1;
			if (osl_keys->pressed.start) menu=0;

			static int debugFrame=0;
			oslPrintf_xy(0,250,"menu=%d and menusel=%d (%d)",menu,menusel,debugFrame++);
		} else {
			//calls the Buttons() function
			ProcessButtons();	
			
			if (osl_keys->pressed.start) menu=1;
		
			//Draw the images to the screen
			oslDrawImage(backgroundImage);
			UpdateMap();
			DrawMap();
			int i;
			for(i=0;i<32;i++) {
				if(koopa[i].health>0) {
					UpdateSprite(koopa+i);
					DrawSprite(koopa+i);
				}
			}
			UpdateSprite(&mario);
			oslSetTransparentColor(RGB(0,255,0));
			DrawSprite(&mario);

			oslPrintf_xy(0,250,"Mario %d,%d (tile %d:%d,%d:%d) mode %d;speed %d, accel %d",mario.x,mario.y, mario.x/16,mario.x%16,mario.y/16,mario.y%16,mario.mode,mario.vertSpeed,mario.vertAccel);
		}

		//Ends drawing mode
		oslEndDrawing();

		//Synchronizes the screen
		oslSyncFrame();		
		
		//Synchronizes the music
		oslAudioVSync();
	}
	
	//Terminate the program
	oslEndGfx();
	oslQuit();
	return 0;
}

void EngineInit()
{
	char buffer[256];
	FILE *file=fopen("Scenery.txt","r");

	printf("Loading in engine settings....\n");
	if(!file) {
		oslDebug("No scenery.txt config file.  That's no fun\n");
	}
	int tagActive=0;
	int frame=0;
	while(fgets(buffer,255,file)) {
		buffer[255]=0;
		if(strchr(buffer,'\r')) strchr(buffer,'\r')[0]=0;
		if(strchr(buffer,'\n')) strchr(buffer,'\n')[0]=0;
		if(buffer[0]=='/') continue;	// skip comments.
		if(buffer[0]=='[') {
			// new tag type
			//printf("Tag count: %d.  Last record: '%c' name: '%s'\n",tileCount, tile[tileCount].shortForm, tile[tileCount].name);
			if(tagActive) tileCount++;
			frame=0;
			tagActive=1;
			if(strchr(buffer,']')) strchr(buffer,']')[0]=0;
			strcpy(tile[tileCount].name,buffer+1);
		} else if(buffer[0]=='{') {
			sscanf(buffer,"{Frame.%d}",&frame);
			if(frame>tile[tileCount].frameCount) {
				tile[tileCount].frameCount=frame;
			}
		} else if(strchr(buffer,'=')) {
			char *name,*value;

			name=buffer;
			value=strchr(buffer,'=');
			value[-1]=0;	// trim a space
			value++;
			value++;
			if(strcmp(name,"ShortForm")==0) {
				tile[tileCount].shortForm=value[0];
				charToTile[(unsigned)value[0]]=tileCount;
			} else if(strcmp(name,"Over")==0) {
				sscanf(value,"%d",&tile[tileCount].x);
				if(frame>0) {
					tile[tileCount].frame[frame-1].x=tile[tileCount].x;
				}
			} else if(strcmp(name,"Down")==0) {
				sscanf(value,"%d",&tile[tileCount].y);
				if(frame>0) {
					tile[tileCount].frame[frame-1].y=tile[tileCount].y;
				}
			} else if(strcmp(name,"Width")==0) {
				sscanf(value,"%d",&tile[tileCount].w);
				if(frame>0) {
					tile[tileCount].frame[frame-1].w=tile[tileCount].w;
				}
			} else if(strcmp(name,"Height")==0) {
				sscanf(value,"%d",&tile[tileCount].h);
				if(frame>0) {
					tile[tileCount].frame[frame-1].h=tile[tileCount].h;
				}
			} else if(strcmp(name,"Type")==0) {
				if(strcmp(value,"SolidGround")==0) {
					tile[tileCount].type=TT_SOLIDGROUND;
				} else if(strcmp(value,"BackGround")==0) {
					tile[tileCount].type=TT_BACKGROUND;
				} else if(strcmp(value,"ItemBlock")==0) {
					tile[tileCount].type=TT_ITEMBLOCK;
				} else if(strcmp(value,"BumpableBlock")==0) {
					tile[tileCount].type=TT_BUMPABLEBLOCK;
				} else if(strcmp(value,"Coin")==0) {
					tile[tileCount].type=TT_COIN;
				} else if(strcmp(value,"DangerBlock")==0) {
					tile[tileCount].type=TT_DANGERBLOCK;
				} else if(strcmp(value,"ExitPipe")==0) {
					tile[tileCount].type=TT_EXITPIPE;
				} else if(strcmp(value,"Climbable")==0) {
					tile[tileCount].type=TT_CLIMBABLE;
				} else {
					oslDebug("Found an unknown tile type in the config file: '%s'.",value);
				}
			} else {
				oslDebug("Found an unknown tile attribute name in the config file: '%s'.",name);
			}
		}
		
	}
	fclose(file);
	
	int i;
	for(i=0;i<SOUNDMAX;i++) {
		sfx[i]=oslLoadSoundFile(soundfilename[i],OSL_FMT_NONE);
		if(!sfx[i]) {
			printf("Couldn't find sound effect '%s'\n",soundfilename[i]);
		}
	}
}

void ShowTiles()
{
	int t=0;
	while (!osl_quit && t<tileCount) {
		//To be able to draw on the screen
		oslStartDrawing();		
		
		//Draw the images to the screen
		oslDrawImage(backgroundImage);
		oslSetTransparentColor(RGB(0,255,0));
		int i,j;
		for(j=0;j<=tile[t].h;j++) {
			for(i=0;i<=tile[t].w;i++) {
				DrawMapTile(32+i*16,64+j*16,tile[t].x+i,tile[t].y+j);
			}
		}
		oslPrintf_xy(10,30,"Tile %d name: '%s'",t,tile[t].name);
		oslPrintf_xy(100,45,"Over,Down: %d,%d",tile[t].x,tile[t].y);
		oslPrintf_xy(100,60,"Width,Height: %d,%d",tile[t].w,tile[t].h);
		oslPrintf_xy(100,75,"Type: %d",tile[t].type);
		oslPrintf_xy(100,75,"ShortForm: %c",tile[t].shortForm);

		//calls the Buttons() function
		oslReadKeys();	
		if(osl_keys->pressed.cross) t++;

		//Ends drawing mode
		oslEndDrawing();

		//Synchronizes the screen
		oslSyncFrame();		
	}
}

int Collision(struct Sprite *sprite,int x,int y)
{
	// First test against map elements
	int tileX=x/16,tileY=y/16;
	int fractX=x%16,fractY=y%16;

	int dirDown=y>sprite->y;
	int dirUp=y<sprite->y;

	int i,j;
	for(j=0;j<sprite->h+(fractY!=0);j++) {
		for(i=0;i<sprite->w+(fractX!=0);i++) {
			int offset=tileX+i+map.width*(tileY+j);
			if(offset<0 || offset>map.width*map.height) return 1;	// never go out of bounds.
			int id=map.tile[offset];
			if(tile[id].type!=TT_CLIMBABLE && sprite->mode==MT_CLIMB) {
				sprite->mode=MT_WALK;
			}
			switch(tile[id].type) {
			case TT_BACKGROUND:
				break;
			case TT_SOLIDGROUND: // can walk on this; has collision all around
				return 1;
			case TT_ITEMBLOCK: // drops powerups
				if(!dirUp) return 1;
				map.map[offset]=TILE(tile[0].x,tile[0].y);
				map.tile[offset]=0;
				// AddItem(tileX,tileY);
				playSound( S_BREAKBLOCK );
				break;
			case TT_BUMPABLEBLOCK: // can destroy the block
				if(!dirUp) return 1;
				map.map[offset]=TILE(tile[0].x,tile[0].y);
				map.tile[offset]=0;
				playSound( S_BREAKBLOCK );
				break;
			case TT_COIN: // mooluah
				if(sprite->type!=ST_PLAYER) break;	// only players can collect coins.
				map.map[offset]=TILE(tile[0].x,tile[0].y);
				map.tile[offset]=0;
				sprite->coins++;
				playSound( S_COINGET );
				break;
			case TT_DANGERBLOCK: // kills on touch
				sprite->health=0;
				playSound( S_MARIODIE );
				break;
			case TT_EXITPIPE: // will exit from the top, but otherwise has collision all around
				if(!dirDown) return 1;
				if(sprite->type!=ST_PLAYER) return 1;
				// teleport to map XXX.
				return 1;
				break;
			case TT_CLIMBABLE:	// can climb up this
				if(sprite->type!=ST_PLAYER) break;
				if(dirUp||dirDown) sprite->mode=MT_CLIMB;
				break;	// can go up, down, left or right.
			}
		}
	}
	return 0;
}

void ProcessButtons()
{
	//initiate the PSP's buttons
	oslReadKeys();

	// a level testing cheat!		
	if(osl_keys->held.triangle) {
		mario.mode=MT_CLIMB;
	} else if(osl_keys->released.triangle) {
		if(mario.mode==MT_CLIMB) mario.mode=MT_WALK;
	}

	if(mario.mode==MT_SWIM || mario.mode==MT_CLIMB) {
		if (osl_keys->held.up) {
			if(!Collision(&mario,mario.x, mario.y-2)) {	
				mario.y -= 2;
			}
		}
		if (osl_keys->held.down) {
			//Sprite movement
			if(!Collision(&mario,mario.x, mario.y+2)) {
				mario.y += 2;
			} else {
				playSound(S_HITWALL);
			}
		}
	}

	if (osl_keys->held.right) {
		if(!Collision(&mario,mario.x+2, mario.y)) {
			mario.x += 2;
		} else {
			playSound(S_HITWALL);
		}
		mario.facing = FACING_RIGHT;
	}
    if (osl_keys->held.left) {
		if(!Collision(&mario,mario.x-2, mario.y)) {
			mario.x -= 2;
		} else {
			playSound(S_HITWALL);
		}
		mario.facing = FACING_LEFT;
	}

	//printf("Handling walk/jump modes\n");
	if(mario.mode==MT_WALK) {
		if(!Collision(&mario,mario.x,mario.y+2)) {
			mario.mode=MT_JUMP;	// we are falling.
			mario.vertSpeed=0;
			printf("starting to fall mario vertSpeed now %d\n",mario.vertSpeed);
			mario.vertAccel=0;
		}
		if(osl_keys->pressed.cross) {
			mario.mode=MT_JUMP;
			playSound(S_JUMP);
			mario.vertSpeed=0;
			printf("starting to jump mario vertSpeed now %d\n",mario.vertSpeed);
			mario.vertAccel=JUMPACCEL;
		}
	} else if(mario.mode==MT_JUMP) {
		if(!osl_keys->held.cross) {
			mario.vertAccel=0;
		}
		if(mario.vertAccel>0 && mario.vertSpeed>=MAXVERTSPEED) {
			mario.vertAccel=0;
		}
		mario.vertSpeed+=mario.vertAccel;
		mario.vertSpeed-=GRAVITY;
		printf("mario vertSpeed now %d\n",mario.vertSpeed);
		if(mario.vertSpeed<MINVERTSPEED) {
			printf("mario vertSpeed %d now MINVERTSPEED\n",mario.vertSpeed);
			mario.vertSpeed=MINVERTSPEED;
		}
		int j;
		for(j=1;j<mario.vertSpeed>>8;j++) {	// moving up in the world.
			if(!Collision(&mario,mario.x,mario.y-1)) {
				mario.y-=1;
				printf("mario moved down 1 to %d\n",mario.y);
			} else {
				mario.vertAccel=0;
				mario.vertSpeed=0;
				printf("bumped head mario vertSpeed now %d\n",mario.vertSpeed);
				playSound(S_HITWALL);
				break;	// bumped head.
			}
		}
		for(j=-1;j>mario.vertSpeed>>8;j--) {	// swept rectangle collision style.
			if(!Collision(&mario,mario.x,mario.y+1)) {
				mario.y+=1;
				printf("mario moved up 1 to %d\n",mario.y);
			} else {
				mario.mode=MT_WALK;
				printf("touched ground mario vertSpeed now %d\n",mario.vertSpeed);
				mario.vertSpeed=0;
				mario.vertAccel=0;
				break;	// touched ground
			}
		}
	}

	//If a button is not pressed
	if (!osl_keys->held.value) 
	{
		//Start the variable over for when a button is pressed again
		mario.frame = 0; 
		mario.frameDelay = 3;
	}
}

void UpdateSprite(struct Sprite *sprite)
{	
	sprite->x+=sprite->speed;	// move at the normal speed.
	if(sprite->x<0) {	// make sure it doesn't go off the edge.
		sprite->x=0;
		sprite->speed=-sprite->speed;
	}
	if(sprite->x>480-16) {	// and the other edge too.
		sprite->x=480-16;
		sprite->speed=-sprite->speed;
	}
	// Slow the frame rate of the walking
	sprite->frameDelay--;
	// Early exit if the delay isn't long enough
	if(sprite->frameDelay>0) return;
	// Reset the delay
	sprite->frameDelay=sprite->desc->delay;

	//Moves the sprite in the row that it is in
	sprite->frame++;
	//resets the sprite movement in that row
	if (sprite->frame >= sprite->desc->frameCount) sprite->frame = 0;
}

void DrawSprite(struct Sprite *sprite)
{
	int top,left;
	left=sprite->desc->frames[sprite->frame].x*17+1; // convert from tiles across to pixels across.
	top=sprite->desc->frames[sprite->frame].y*17+1;	// convert from tiles down to pixels down.
	oslSetImageTileSize(sprite->spriteImage,left,top,16,16);	// set the sprite sheet up to show that sprite.
	sprite->spriteImage->x=sprite->x;	// draw it where it goes
	sprite->spriteImage->y=sprite->y;	// draw it where it goes
	if(sprite->facing==FACING_LEFT) oslMirrorImageH(sprite->spriteImage);	// flip
	oslDrawImage(sprite->spriteImage);
	if(sprite->facing==FACING_LEFT) oslMirrorImageH(sprite->spriteImage);	// return to normal.
}

void NewKoopa(int x,int y)
{
	int i;
	for(i=0;i<32;i++) {
		if(koopa[i].health==0) {
			koopa[i].type=ST_ENEMY;
			koopa[i].mode=MT_WALK;
			koopa[i].spriteImage=enemiesImage;
			koopa[i].health=1;
			koopa[i].x=x;
			koopa[i].y=y;
			koopa[i].speed=-2;
			koopa[i].facing=FACING_LEFT;
			koopa[i].frame=0;
			koopa[i].frameDelay=3;
			koopa[i].w=1;
			koopa[i].h=1;
			koopa[i].vertSpeed=0;
			koopa[i].vertAccel=0;
			koopa[i].desc=&koopaWalkDesc;
			koopa[i].coins=0;
		}
	}
}

// translate a tile in a level file, into a tile in the sprite sheet
int mapCharToTile( char ch)
{
	int id;

	switch(ch) {
	//case ' ': return TILE(29,9);	// space
	case 'u': return ENEM(25,4);	// ghost
	case 'v': return ENEM(7,2);	// bomb
	case 'x': return ENEM(3,3);	// turtle
	case 'y': return ENEM(1,2);	// koopa
	case 'z': return ENEM(20,0);	// bear
	default:
		// still doesn't quite handle multitile objects.
		id=charToTile[(int)ch];
		printf("matched tile '%c'to id %d (%s)\n",ch,id,tile[id].name);
		return TILE(tile[id].x,tile[id].y);
	}
}

void LoadMap(const char *filename)
{
	char line[256];
	printf("Locating map '%s'\n",filename);
	FILE *file = fopen( filename, "r");
	if(!file) return;
	printf("Loading map...\n");
	if(map.map) free(map.map);
	if(map.tile) free(map.tile);
	if(map.animTile) free(map.animTile);
	map.width=30;	// PSP screen is 480/16
	map.height=17;	// PSP screen is 272/16
	map.map = (int *)calloc( sizeof(int), map.width*map.height);
	map.tile = (int *)calloc( sizeof(int), map.width*map.height);
	map.animTile = (struct AnimTile *)calloc( sizeof(struct AnimTile), 1024);
	map.animTileCount = 0;
	int j=0;
	int i;
	while( fgets( line, 255, file) ) {
		line[255] = 0;
		if(strrchr( line, '\n') ) strrchr( line, '\n')[0] = 0;
		for(i = 0; i<strlen(line); i++) {
			if( i < map.width) {
				// Set default tile for the map as an open square.
				map.map[i+map.width*j]=TILE(29,9);
				map.tile[i+map.width*j]=0;	// sky
				// Handle special tiles, for active enemies, and Mario's start point.
				switch(line[i]) {
				case 'M':	// mario start point
					mario.spriteImage=playerImage;
					mario.x=i*16;
					mario.y=j*16;
					mario.facing=FACING_RIGHT;
					mario.health=1;
					mario.speed=0;
					mario.frame=0;
					mario.frameDelay=3;
					mario.w=1;
					mario.h=1;
					mario.desc=&marioWalkDesc;
					mario.vertSpeed=0;
					mario.vertAccel=0;
					break;
				case 'y':	// koopa
					NewKoopa(i*16,j*16);
					break;
				default:
					{
						int tileNo=mapCharToTile(line[i]);
						int tileId=charToTile[(unsigned char)line[i]];
						map.tile[i+map.width*j]=tileId;
						if(tile[tileId].w>0 || tile[tileId].h>0) {
							if(i>0) {
								int tileLeft=map.map[(i-1)+map.width*j];
								int xx=tileLeft%32;
								int yy=tileLeft/32;
								if(xx>=tile[tileId].x && xx<=tile[tileId].x+tile[tileId].w
								&& yy>=tile[tileId].y && yy<=tile[tileId].y+tile[tileId].h) {
									int ox=xx-tile[tileId].x;
									int oy=yy-tile[tileId].y;
									ox+=1;
									if(ox>tile[tileId].w) ox=0;
									tileNo=mapCharToTile(line[i])+TILE(ox,oy);
								}
							}
							if(j>0) {
								int tileAbove=map.map[i+map.width*(j-1)];
								int xx=tileAbove%32;
								int yy=tileAbove/32;
								if(xx>=tile[tileId].x && xx<=tile[tileId].x+tile[tileId].w
								&& yy>=tile[tileId].y && yy<=tile[tileId].y+tile[tileId].h) {
									int ox=xx-tile[tileId].x;
									int oy=yy-tile[tileId].y;
									oy+=1;
									if(oy>tile[tileId].h) oy=0;
									tileNo=mapCharToTile(line[i])+TILE(ox,oy);
								}
							}
						}
						map.map[i+map.width*j]=tileNo;
					}
				}
				if(tile[map.tile[i+map.width*j]].frameCount>1) {
					if(map.animTileCount%1024==1023) {
						map.animTile=realloc(map.animTile,sizeof(struct AnimTile)*(map.animTileCount+1025));
					}
					map.animTile[map.animTileCount].x=i;
					map.animTile[map.animTileCount].y=j;
					map.animTile[map.animTileCount].frame=0;
					map.animTile[map.animTileCount].tile=tile+map.tile[i+map.width*j];
				}
			}
		}
		j++;
		if( j >= map.height) break;

	}
	fclose(file);
	printf("Loaded map\n");
	playSound(S_MARIOSTART);
}

void UpdateMap()
{
	static int timer=6;
	timer--;
	if(timer>0) return;
	timer=6;
	int i;
	for(i=0;i<map.animTileCount;i++) {
		struct AnimTile *animTile=&map.animTile[i];
		animTile->frame++;
		if(animTile->frame>=animTile->tile->frameCount) animTile->frame=0;
		struct TilePos *tp=&animTile->tile->frame[animTile->frame];
		// I don't currently handle animated tiles that are bigger than 1x1
		map.map[animTile->x+animTile->y*map.width]=TILE(tp->x,tp->y);
	}
}

void DrawMapTile(int screenX,int screenY,int tileX,int tileY)
{
	OSL_IMAGE *spriteImage;

	if(tileY>63) {
		spriteImage=enemiesImage;
		tileY-=64;
	} else if(tileY>31) {
		spriteImage=itemsImage;
		tileY-=32;
	} else {
		spriteImage=sceneryImage;
	}
	int left=tileX*17+1;	// convert from tiles across to pixels across.
	int top=tileY*17+1;	// convert from tiles down to pixels down.
	oslSetImageTileSize(spriteImage,left,top,16,16);	// set the sprite sheet up to show that sprite.
	spriteImage->x=screenX;	// draw it where it goes
	spriteImage->y=screenY;	// draw it where it goes
	oslDrawImage(spriteImage);
}

void DrawMap()
{
	int i,j;
	if(!map.map) return;
	for(j=0; j<map.height; j++) {
		for(i=0; i<map.width; i++) {
			int x,y;
			int tile=map.map[i+j*map.width];
			if(tile==0) tile=TILE(29,9);	// great for incomplete maps.
			x=tile&31;
			y=(tile>>5)&127;
			DrawMapTile(i*16,j*16,x,y);
		}
	}
}

void playSound( int sound)
{
	static int channel=3;
	
	int chan=channel;
	if(sound==S_COINGET) channel=1;
	if(sound==S_MARIODIE) channel=2;
	if(sound==S_HITWALL) channel=2;
	if(sound>=0 && sound<SOUNDMAX && sfx[sound]) {
		oslPlaySound( sfx[sound], channel);
	} else {
		printf("Sound %d missing.\n",sound);
	}
	if(chan==channel) channel++;
	if(channel>3) channel=1;
}
